
/******************************************************************************
* MODULE     : table.hpp
* DESCRIPTION: Hash tables
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_TABLE_HPP
#define __MML_TABLE_HPP
#include "list.hpp"
namespace mml {
#define TMPL_DEF template<typename C, typename T>
#define TMPL template<typename C, typename T>
#define Table table<C,T>
#define Table_rep table_rep<C,T>
TMPL class table_rep;
TMPL class table;
TMPL class entries_iterator_rep;
TMPL class table_iterator_rep;
TMPL C& ref_I (Table& t);
TMPL C I (const Table& t);
TMPL nat N (const Table& t);
TMPL void reset (Table& t, const T& x);
TMPL void simplify (Table& t);
TMPL bool operator == (const Table& t1, const Table& t2);

/******************************************************************************
* Pairs
******************************************************************************/

template<typename T, typename C>
class pair {
public:
  T key;
  C im;
  pair () {}
  pair (const T& key2, const C& im2): key (key2), im (im2) {}
};

template<typename T, typename C> nat
hash (const pair<T,C>& p) {
  nat h= hash (p.key);
  return (h<<1) ^ (h<<3) ^ (h<<9) ^ (h>>29) ^ hash (p.im);
}

template<typename T, typename C> ostream&
operator << (ostream& out, const pair<T,C>& p) {
  out << p.key << "->" << p.im;
  return out;
}

template<typename T, typename C> content
as_content (const pair<T,C>& p) {
  return as_content ("->", p.key, p.im);
}

template<typename T, typename C> bool
operator == (const pair<T,C>& p1, const pair<T,C>& p2) {
  return (p1.key==p2.key) && (p1.im==p2.im);
}

template<typename T, typename C> bool
operator != (const pair<T,C>& p1, const pair<T,C>& p2) {
  return (p1.key!=p2.key) || (p1.im!=p2.im);
}

/******************************************************************************
* Table class and its representation class
******************************************************************************/

TMPL_DEF
class table_rep REP_STRUCT {
  nat size;             // size of table (nr of entries)
  nat n;                // nr of keys (a power of two)
  C   init;             // default entry
  list<pair<T,C> >* a;  // the array of entries

public:
  Table_rep (const C& init2, nat n2=1):
    size(0), n(n2), init(init2), a(new list<pair<T,C> >[n]) {}
  ~Table_rep () { delete[] a; }
  void resize (nat n);
  void reset (const T& x);
  bool contains (const T& x) const;
  C    bracket_ro (const T& x) const;
  C&   bracket_rw (const T& x);
  void simplify ();

  friend class Table;
  friend class entries_iterator_rep<C,T>;
  friend class table_iterator_rep<C,T>;
  friend C& ref_I LESSGTR (Table& t);
  friend C I LESSGTR (const Table& t);
  friend nat N LESSGTR (const Table& t);
  friend bool operator == LESSGTR (const Table& t1, const Table& t2);
};

TMPL_DEF
class table {
INDIRECT_PROTO_2 (table, table_rep, C, T)
  table (): rep (new Table_rep (C(), 1)) {}
  table (int i): rep (new Table_rep (C(i), 1)) {}
  table (const C& init, nat n=1): rep (new Table_rep (init, n)) {}
  table (const iterator<pair<T,C> >& it): rep (new Table_rep (C(), 1)) {
    for (; busy (it); ++it) rep->bracket_rw ((*it).key)= (*it).im; }
  C  operator [] (const T& x) const { return rep->bracket_ro (x); }
  C& operator () (const T& x) { secure (); return rep->bracket_rw (x); }
  friend C& ref_I LESSGTR (Table& t);
  friend void reset LESSGTR (Table& t, const T& x);
  friend void simplify LESSGTR (Table& t);
};
INDIRECT_IMPL_2 (table, table_rep, typename C, C, typename T, T)

TMPL void reset (Table& t, const T& x) { t.secure(); t.rep->reset (x); }
TMPL void simplify (Table& t) { t.secure(); t.rep->simplify (); }

/******************************************************************************
* Basic access routines for tables
******************************************************************************/

TMPL C&
ref_I (Table& t) {
  return t.rep->init;
}

TMPL C
I (const Table& t) {
  return t->init;
}

TMPL nat
N (const Table& t) {
  return t->size;
}

TMPL void
Table_rep::resize (nat n2) {
  nat i, oldn= n;
  list<pair<T,C> >* olda= a;
  n= n2;
  a= new list<pair<T,C> >[n];
  for (i=0; i<oldn; i++) {
    list<pair<T,C> > l (olda [i]);
    //cout << "  [" << i << ", " << N(l) << "]\n";
    while (!nil (l)) {
      nat code= hash (car(l).key);
      list<pair<T,C> >& newl= a [code & (n-1)];
      newl= cons (car (l), newl);
      l= cdr (l);
    }
  }
  delete[] olda;
  //double acc= 0.0;
  //for (i=0; i<n; i++)
  //acc += N(a[i]) * N(a[i]);
  //cout << "Access ratio: " << (acc/n) << " out of " << n << "\n";
}

TMPL bool
Table_rep::contains (const T& x) const {
  list<pair<T,C> >  l (a [hash (x) & (n-1)]);
  while (!nil (l)) {
    if (car (l).key==x) return true;
    l= cdr (l);
  }
  return false;
}

TMPL C&
Table_rep::bracket_rw (const T& x) {
  register nat hv= hash(x);
  list<pair<T,C> >* l= &(a [hv & (n-1)]);
  while (!nil (*l)) {
    if (car (*l).key == x)
      return ref_car (*l).im;
    l= &ref_cdr (*l);
  }
  //cout << "size= " << size << ", n= " << n << "\n";
  if (size>=n) resize (n<<1);
  //cout << "Resized\n";
  l= &a [hv & (n-1)];
  *l= cons (pair<T,C> (x, init), *l);
  size ++;
  return ref_car (*l).im;
}

TMPL C
Table_rep::bracket_ro (const T& x) const {
  list<pair<T,C> > l (a [hash (x) & (n-1)]);
  while (!nil (l)) {
    if (car (l).key == x) return car (l).im;
    l= cdr (l);
  }
  return init;
}

TMPL void
Table_rep::reset (const T& x) {
  list<pair<T,C> > *l= &(a [hash (x) & (n-1)]);
  while (!nil (*l)) {
    if (car (*l).key == x) {
      *l= ref_cdr (*l);
      size --;
      if (size < (n>>1)) resize (n>>1);
      return;
    }
    l= &ref_cdr (*l);
  }
}

TMPL void
Table_rep::simplify () {
  for (nat i=0; i<n; i++) {
    list<pair<T,C> > *l= &(a[i]);
    while (!nil (*l)) {
      if (car (*l).im == init) {
	*l= ref_cdr (*l);
	size --;
      }
      else l= &ref_cdr (*l);
    }
  }
  while (size < (n>>1)) resize (n>>1);
}

/******************************************************************************
* Iterators
******************************************************************************/

TMPL_DEF
class entries_iterator_rep: public iterator_rep<T> {
  Table t;             // the table to traverse
  nat i;               // current entry in table
  list<pair<T,C> > l;  // current list

protected:
  void spool () {
    while (nil (l)) {
      i++;
      if (i >= t->n) break;
      l= t->a[i];
    }
  }
  bool is_busy () { return i<t->n; }
  void advance () { l= cdr (l); spool (); }
  T current () { return car (l).key; }

public:
  entries_iterator_rep (const Table& t2):
    t(t2), i(0), l(t->a[0]) { spool (); }
};

TMPL iterator<T>
entries (const Table& t) {
  return iterator<T> (new entries_iterator_rep<C,T> (t));
}

TMPL_DEF
class table_iterator_rep: public iterator_rep<pair<T,C> > {
  Table t;        // the table to traverse
  nat i;               // current entry in table
  list<pair<T,C> > l;  // current list

protected:
  void spool () {
    while (nil (l)) {
      i++;
      if (i >= t->n) break;
      l= t->a[i];
    }
  }
  bool is_busy () { return i<t->n; }
  void advance () { l= cdr (l); spool (); }
  pair<T,C> current () { return car (l); }

public:
  table_iterator_rep (const Table& t2):
    t(t2), i(0), l(t->a[0]) { spool (); }
};

TMPL iterator<pair<T,C> >
iterate (const Table& t) {
  return iterator<pair<T,C> > (new table_iterator_rep<C,T> (t));
}

/******************************************************************************
* Hashing and outputting tables
******************************************************************************/

TMPL nat
hash (const Table& t) {
  // we add all hash codes of the element pairs; don't "cycle",
  // because the order of the element pairs is not uniquely determined
  nat h=54321;
  for (iterator<pair<T,C> > it= iterate (t); busy (it); ++it)
    h += hash (*it);
  return h;
}

TMPL content
as_content (const Table& t) {
  content args;
  for (iterator<pair<T,C> > it= iterate (t); busy (it); ++it)
    args << as_content (*it);
  return apply ("table", args);
}

PRINT_SUGAR(TMPL,Table)

/******************************************************************************
* Vectorial operations on tables
******************************************************************************/

TMPL Table
copy (const Table& t) {
  Table r (I(t));
  for (iterator<T> it= entries (t); busy (it); ++it)
    r(*it)= t[*it];
  return r;
}

#define TABLE_UNARY(op)					\
TMPL Table						\
operator op (const Table& t) {				\
  Table r (I(t));					\
  for (iterator<T> it= entries (t); busy (it); ++it)	\
    r(*it)= op t[*it];					\
  simplify (r);						\
  return r;						\
}

#define TABLE_BINARY(op)				\
TMPL Table						\
operator op (const Table& t, const Table& u) {		\
  Table r (I(t) op I(u));				\
  for (iterator<T> it= entries (t); busy (it); ++it)	\
    r(*it)= t[*it] op u[*it];				\
  for (iterator<T> it= entries (u); busy (it); ++it)	\
    if (!(t->contains (*it)))				\
      r(*it)= t[*it] op u[*it];				\
  simplify (r);						\
  return r;						\
}

#define TABLE_SCALAR(op)				\
TMPL Table						\
operator op (const Table& t, const C& c) {		\
  Table r (I(t));					\
  for (iterator<T> it= entries (t); busy (it); ++it)	\
    r(*it)= t[*it] op c;				\
  simplify (r);						\
  return r;						\
}

#define TABLE_INLINE_BINARY(op)				\
TMPL Table&						\
operator op (Table& t, const Table& u) {		\
  ref_I(t) op I(u);					\
  for (iterator<T> it= entries (u); busy (it); ++it) {	\
    t(*it) op u[*it];					\
    if (t[*it] == I(t)) reset (t, *it);			\
  }							\
  return t;						\
}

#define TABLE_INLINE_SCALAR(op)				\
TMPL Table&						\
operator op (Table& t, const C& c) {			\
  ref_I(t) op c;					\
  for (iterator<T> it= entries (t); busy (it); ++it)	\
    t(*it) op c;					\
  simplify (t);						\
  return t;						\
}

TABLE_UNARY(-);
TABLE_BINARY(+);
TABLE_BINARY(-);
TABLE_BINARY(*);
TABLE_BINARY(/);
TABLE_BINARY(%);
TABLE_SCALAR(*);
TABLE_SCALAR(/);
TABLE_SCALAR(%);
TABLE_INLINE_BINARY(+=);
TABLE_INLINE_BINARY(-=);
TABLE_INLINE_BINARY(*=);
TABLE_INLINE_BINARY(/=);
TABLE_INLINE_BINARY(%=);
TABLE_INLINE_SCALAR(*=);
TABLE_INLINE_SCALAR(/=);
TABLE_INLINE_SCALAR(%=);

TMPL Table
operator * (const C& c, const Table& t) {
  return t * c;
}

/******************************************************************************
* Comparison routines for tables
******************************************************************************/

#define TABLE_TEST(op)							\
TMPL bool								\
operator op (const Table& t, const Table& u) {				\
  if (!(I(t) op I(u))) return false;					\
  for (iterator<T> it= entries (t); busy (it); ++it)			\
    if (!(t[*it] op u[*it])) return false;				\
  for (iterator<T> it= entries (u); busy (it); ++it)			\
    if (!(t->contains (*it) || (t[*it] op u[*it]))) return false;	\
  return true;								\
}

TABLE_TEST(==);
TABLE_TEST(<=);
TABLE_TEST(>=);

/******************************************************************************
* Syntactic sugar
******************************************************************************/

ARITH_INT_SUGAR(TMPL,Table,C);
SHIFT_INT_SUGAR(TMPL,Table,C);
STRICT_COMPARE_SUGAR(TMPL,Table)

#undef TMPL_DEF
#undef TMPL
#undef Table
#undef Table_rep
} // namespace mml
#endif // __MML_TABLE_HPP
