
/******************************************************************************
* MODULE     : real.hpp
* DESCRIPTION: Effective real numbers
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_REAL_HPP
#define __MML_REAL_HPP
#include "dyadic.hpp"
#include "interval.hpp"
namespace mml {

#define MAX_CANCEL_TEST mml_default_zt

class real_rep REP_STRUCT {
public:
  const dyadic wt; // weight of real number as an expression
  interval best;   // the best currently available approximation
                   // FIXME: best should be protected
public:
  inline real_rep (const dyadic& wt2):
    wt (wt2), best (interval::fuzzy) {}
  inline real_rep (const dyadic& wt2, const interval& best2):
    wt (wt2), best (best2) {}
  virtual inline ~real_rep () {}

  virtual content expression () const = 0;
  virtual dyadic get_bound () = 0;
  virtual interval approx (const dyadic& err) = 0;

  mp_prec_t linear_algo (const dyadic& err);
  mp_prec_t mult_algo (const dyadic& err);
  bool normalize (mp_prec_t max_cancel, const dyadic& rel_err);
  interval expand (mp_prec_t prec);
};

class real {
INDIRECT_PROTO (real, real_rep)
public:
  real (signed int i);
  real (unsigned int i);
  real (signed short int i);
  real (unsigned short int i);
  real (signed long int i);
  real (unsigned long int i);

  inline friend dyadic bound (const real& x);
  inline friend interval approximate (const real& x, const dyadic& err) {
    return x.rep->approx (err); }
  inline friend interval
  expand (const real& x, mp_prec_t prec= mml_default_prec) {
    return x.rep->expand (prec); }
  inline friend bool
  normalize (const real& x, mp_prec_t max_cancel= MAX_CANCEL_TEST,
	     const dyadic& rel_err= error (1, -MIN_SECURITY)) {
    return x.rep->normalize (max_cancel, rel_err); }
};
INDIRECT_IMPL (real, real_rep)

// The precision of a computed approximation can be a little bit too low
// due to rounding errors. In that case, we increase the precision.
// If the computed precision is far too low, then this probably indicates
// a bug in the code.
#define ENSURE_REAL_CONSISTENCY(err)				\
  if (wider (best, err)) {					\
    dyadic factor= radius (best) / err;				\
    if (factor > 8) {						\
      cerr << "\nPrecision deficiency: " << factor << "\n";	\
      assert (factor <= 8);					\
    }								\
    return approx (err / (factor + 1));				\
  }

inline dyadic weight (const real& x1) {
  return x1->wt; }
inline dyadic weight (const real& x1, const real& x2) {
  return add_up (x1->wt, x2->wt); }

content as_content (const real& x);
inline ostream& operator << (ostream& out, const real& x) {
  return out << as_content (x); }
inline dyadic bound (const real& x) {
  if (is_normal (x.rep->best)) return bound (x->best);
  else return x.rep->get_bound (); }

real operator - (const real& x);
real operator + (const real& x, const real& y);
real operator - (const real& x, const real& y);
real operator * (const real& x, const real& y);
real operator / (const real& x, const real& y);
real operator % (const real& x, const real& y);
real operator << (const real& x, mp_exp_t shift);
real operator >> (const real& x, mp_exp_t shift);

real abs (const real& x);
int sign (const real& x, int max_cancel= MAX_CANCEL_TEST);

ARITH_INT_SUGAR(inline,real,real);
SHIFT_INT_SUGAR(inline,real,real);
COMPARE_SUGAR(inline,real);

}
#endif // __MML_REAL_HPP
