
/******************************************************************************
* MODULE     : symbolic.hpp
* DESCRIPTION: Symbolic expressions
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_HPP
#define __SYMBOLIC_HPP
#include "string.hpp"
#include "dense_vector.hpp"
#include "symbolic_defs.hpp"
namespace mml {

class symbolic;
class symbolic_aux;
class context;

class symbolic_rep: public rep_struct {
public: // FIXME: should become protected
  // accessors and syntactic identity
  virtual symbolic_type type () const = 0;
  virtual context get_context () const;
  virtual nat hash () const = 0;
  virtual bool identical (symbolic_rep* rep) const = 0;
  virtual symbolic flatten (bool recursive) const = 0;
  virtual bool is_atomic ();
  virtual string leaf ();
  virtual nat arity ();
  virtual symbolic child (nat i);

  // properties
  virtual add_type get_add_type () const;  // type w.r.t. addition
  virtual mul_type get_mul_type () const;  // type w.r.t. multiplication
  virtual pow_type get_pow_type () const;  // type w.r.t. taking powers

public:
  inline symbolic_rep () {}
  inline virtual ~symbolic_rep () {}
  bool in (const context& c) const;
  friend class symbolic;
};

class symbolic {
protected:
  symbolic_rep* rep;

public:
  inline symbolic (const symbolic& x);
  inline const symbolic_rep* operator -> () const;
  symbolic (symbolic_rep* rep);
  symbolic& operator = (const symbolic& x);
  ~symbolic ();

public:
  symbolic ();
  symbolic (int i);
  symbolic (const integer& i);
  symbolic (const rational& r);
  symbolic (char* s);
  symbolic (const string& s);

  friend integer as_integer  (const symbolic& s);
  friend rational as_rational (const symbolic& s);
  inline friend bool is_atomic (const symbolic& s);
  inline friend string L (const symbolic& s);
  inline friend bool is_compound (const symbolic& s);
  inline friend nat N (const symbolic& s);
  inline symbolic operator [] (nat i) const;
  inline friend nat hash (const symbolic& s);
  inline friend symbolic flatten (const symbolic& s);
  inline friend symbolic flatten (const symbolic& s, bool really);
  inline friend const symbolic_rep* access (const symbolic& s);
  inline bool operator == (const symbolic& s) const;
  inline bool operator != (const symbolic& s) const;
  inline bool operator <  (const symbolic& s) const ;
  inline bool operator <= (const symbolic& s) const;
  inline bool operator >  (const symbolic& s) const;
  inline bool operator >= (const symbolic& s) const;
  inline symbolic operator () ();
  inline symbolic operator () (const symbolic& s1);
  inline symbolic operator () (const symbolic& s1, const symbolic& s2);
  inline symbolic operator () (const vector<symbolic>& args);
  friend ostream& operator << (ostream& out, const symbolic& s);

  friend class symbolic_rep;
};

inline symbolic::symbolic (const symbolic& x): rep(x.rep) { INC_COUNT (rep); }
inline const symbolic_rep* symbolic::operator -> () const { return rep; }
inline bool is_atomic (const symbolic& s) { return s.rep->is_atomic (); }
inline string L (const symbolic& s) { return s.rep->leaf (); }
inline bool is_compound (const symbolic& s) { return !s.rep->is_atomic (); }
inline nat N (const symbolic& s) { return s.rep->arity (); }
inline symbolic symbolic::operator [] (nat i) const { return rep->child (i); }
inline nat hash (const symbolic& s) { return ((nat) s.rep) >> 2; }
inline symbolic flatten (const symbolic& s) { return s.rep->flatten (true); }
inline symbolic flatten (const symbolic& s, bool really) {
  return really? s.rep->flatten (true): s; }
inline const symbolic_rep* access (const symbolic& s) { return s.rep; }

inline bool symbolic::operator == (const symbolic& s) const {
  return rep == s.rep; }
inline bool symbolic::operator != (const symbolic& s) const {
  return rep != s.rep; }
inline bool symbolic::operator <  (const symbolic& s) const {
  return rep <  s.rep; }
inline bool symbolic::operator <= (const symbolic& s) const {
  return rep <= s.rep; }
inline bool symbolic::operator >  (const symbolic& s) const {
  return rep >  s.rep; }
inline bool symbolic::operator >= (const symbolic& s) const {
  return rep >= s.rep; }

void lisp_print (ostream& out, const symbolic& s);
void pretty_print (ostream& out, const symbolic& s);

symbolic symbolic_compound (const vector<symbolic>& fun_args);
symbolic symbolic_compound (const symbolic& fun, const vector<symbolic>& args);
symbolic symbolic_compound (const symbolic& fun);
symbolic symbolic_compound (const symbolic& fun, const symbolic& s1);
symbolic symbolic_compound (const symbolic& fun,
			    const symbolic& s1, const symbolic& s2);

symbolic apply (const vector<symbolic>& fun_args);
symbolic apply (const symbolic& fun, const vector<symbolic>& args);
symbolic apply (const symbolic& fun);
symbolic apply (const symbolic& fun, const symbolic& s1);
symbolic apply (const symbolic& fun, const symbolic& s1, const symbolic& s2);

inline symbolic symbolic::operator() () {
  return apply (*this); }
inline symbolic symbolic::operator() (const symbolic& s1) {
  return apply (*this, s1); }
inline symbolic symbolic::operator() (const symbolic& s1, const symbolic& s2) {
  return apply (*this, s1, s2); }
inline symbolic symbolic::operator() (const vector<symbolic>& args) {
  return apply (*this, args); }

symbolic operator - (const symbolic& s1);
symbolic operator + (const symbolic& s1, const symbolic& s2);
symbolic operator - (const symbolic& s1, const symbolic& s2);
symbolic operator * (const symbolic& s1, const symbolic& s2);
symbolic operator / (const symbolic& s1, const symbolic& s2);
symbolic operator % (const symbolic& s1, const symbolic& s2);
symbolic pow (const symbolic& s1, const symbolic& s2);

} // namespace mml
#endif // __SYMBOLIC_HPP
