
/******************************************************************************
* MODULE     : context_base.cpp
* DESCRIPTION: Common utilities for contexts for symbolic expressions
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "context_base.hpp"
namespace mml {

/******************************************************************************
* Default dispatching
******************************************************************************/

#define MAKE(what) evaluator ((evaluator_rep*) new what (me ()))

struct compound_evaluator_rep: public evaluator {
  context c;
  compound_evaluator_rep (const context& c2): c (c2) {}
  symbolic eval (const symbolic& s) { return c->eval_compound (s); }
  content expression () const { return SYMB_APPLY; }
};

struct add_evaluator_rep: public evaluator {
  context c;
  add_evaluator_rep (const context& c2): c (c2) {}
  symbolic eval (const symbolic& s) {
    if (N(s) == 2) return c->add (s[1], s[2]);
    else return c->eval_compound (s); }
  content expression () const { return SYMB_PLUS; }
};

struct sub_evaluator_rep: public evaluator {
  context c;
  sub_evaluator_rep (const context& c2): c (c2) {}
  symbolic eval (const symbolic& s) {
    if (N(s) == 1) return c->neg (s[1]);
    else if (N(s) == 2) return c->sub (s[1], s[2]);
    else return c->eval_compound (s); }
  content expression () const { return SYMB_MINUS; }
};

struct mul_evaluator_rep: public evaluator {
  context c;
  mul_evaluator_rep (const context& c2): c (c2) {}
  symbolic eval (const symbolic& s) {
    if (N(s) == 2) return c->mul (s[1], s[2]);
    else return c->eval_compound (s); }
  content expression () const { return SYMB_TIMES; }
};

struct div_evaluator_rep: public evaluator {
  context c;
  div_evaluator_rep (const context& c2): c (c2) {}
  symbolic eval (const symbolic& s) {
    if (N(s) == 2) return c->div (s[1], s[2]);
    else return c->eval_compound (s); }
  content expression () const { return SYMB_OVER; }
};

struct pow_evaluator_rep: public evaluator {
  context c;
  pow_evaluator_rep (const context& c2): c (c2) {}
  symbolic eval (const symbolic& s) {
    if (N(s) == 2) return c->pow (s[1], s[2]);
    else return c->eval_compound (s); }
  content expression () const { return SYMB_POWER; }
};

/******************************************************************************
* Methods for context_base_rep
******************************************************************************/

context_base_rep::context_base_rep (const context& base2, const symbolic& id2):
  context_rep (id2), base (base2), lcc (context ()),
  dispatch (MAKE (compound_evaluator_rep))
{
  dispatch (SYMB_PLUS) = MAKE (add_evaluator_rep);
  dispatch (SYMB_MINUS)= MAKE (sub_evaluator_rep);
  dispatch (SYMB_TIMES)= MAKE (mul_evaluator_rep);
  dispatch (SYMB_OVER) = MAKE (div_evaluator_rep);
  dispatch (SYMB_POWER)= MAKE (pow_evaluator_rep);
}

bool
context_base_rep::extends (const context& c) const {
  return c == me () || base->extends (c);
}

context
context_base_rep::least_common_context (const context& c) {
  if (extends (c)) return me ();
  assert (c->extends (me ()));
  return c;
}

context
context_base_rep::fetch_lcc (const context& c) {
  if (c == context ()) return me ();
  assert (c->extends (me ()));
  return c;
}

symbolic
context_base_rep::eval (const symbolic& s) const {
  if (is_atomic (s)) return eval_atomic (s);
  else return dispatch [s[0]] (s);
}

symbolic
context_base_rep::eval_atomic (const symbolic& s) const {
  return base->eval_atomic (s);
}

symbolic
context_base_rep::eval_compound (const symbolic& s) const {
  return base->eval_compound (s);
}

/******************************************************************************
* Common operations
******************************************************************************/

symbolic
context_base_rep::add (const symbolic& s1, const symbolic& s2) const {
  return base->add (s1, s2);
}

symbolic
context_base_rep::sub (const symbolic& s1, const symbolic& s2) const {
  return base->sub (s1, s2);
}

symbolic
context_base_rep::neg (const symbolic& s1) const {
  return base->neg (s1);
}

symbolic
context_base_rep::mul (const symbolic& s1, const symbolic& s2) const {
  return base->mul (s1, s2);
}

symbolic
context_base_rep::div (const symbolic& s1, const symbolic& s2) const {
  return base->div (s1, s2);
}

symbolic
context_base_rep::pow (const symbolic& s1, const symbolic& s2) const {
  return base->pow (s1, s2);
}

} // namespace mml
